/****************************************************************************
** Meta object code from reading C++ file 'ui_qt.h'
**
** Created by: The Qt Meta Object Compiler version 69 (Qt 6.11.1)
**
** WARNING! All changes made in this file will be lost!
*****************************************************************************/

#include "../../../../ui/drivers/ui_qt.h"
#include <QtCore/qmetatype.h>
#include <QtCore/QList>

#include <QtCore/qtmochelpers.h>

#include <memory>


#include <QtCore/qxptype_traits.h>
#if !defined(Q_MOC_OUTPUT_REVISION)
#error "The header file 'ui_qt.h' doesn't include <QObject>."
#elif Q_MOC_OUTPUT_REVISION != 69
#error "This file was generated using the moc from 6.11.1. It"
#error "cannot be used with the include files from this version of Qt."
#error "(The moc has changed too much.)"
#endif

#ifndef Q_CONSTINIT
#define Q_CONSTINIT
#endif

QT_WARNING_PUSH
QT_WARNING_DISABLE_DEPRECATED
QT_WARNING_DISABLE_GCC("-Wuseless-cast")
namespace {
struct qt_meta_tag_ZN13PlaylistModelE_t {};
} // unnamed namespace

template <> constexpr inline auto PlaylistModel::qt_create_metaobjectdata<qt_meta_tag_ZN13PlaylistModelE_t>()
{
    namespace QMC = QtMocConstants;
    QtMocHelpers::StringRefStorage qt_stringData {
        "PlaylistModel",
        "imageLoaded",
        "",
        "QImage",
        "image",
        "QModelIndex",
        "index",
        "path",
        "onImageLoaded"
    };

    QtMocHelpers::UintData qt_methods {
        // Signal 'imageLoaded'
        QtMocHelpers::SignalData<void(const QImage, const QModelIndex &, const QString &)>(1, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 3, 4 }, { 0x80000000 | 5, 6 }, { QMetaType::QString, 7 },
        }}),
        // Slot 'onImageLoaded'
        QtMocHelpers::SlotData<void(const QImage, const QModelIndex &, const QString &)>(8, 2, QMC::AccessPrivate, QMetaType::Void, {{
            { 0x80000000 | 3, 4 }, { 0x80000000 | 5, 6 }, { QMetaType::QString, 7 },
        }}),
    };
    QtMocHelpers::UintData qt_properties {
    };
    QtMocHelpers::UintData qt_enums {
    };
    return QtMocHelpers::metaObjectData<PlaylistModel, qt_meta_tag_ZN13PlaylistModelE_t>(QMC::MetaObjectFlag{}, qt_stringData,
            qt_methods, qt_properties, qt_enums);
}
Q_CONSTINIT const QMetaObject PlaylistModel::staticMetaObject = { {
    QMetaObject::SuperData::link<QAbstractListModel::staticMetaObject>(),
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN13PlaylistModelE_t>.stringdata,
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN13PlaylistModelE_t>.data,
    qt_static_metacall,
    nullptr,
    qt_staticMetaObjectRelocatingContent<qt_meta_tag_ZN13PlaylistModelE_t>.metaTypes,
    nullptr
} };

void PlaylistModel::qt_static_metacall(QObject *_o, QMetaObject::Call _c, int _id, void **_a)
{
    auto *_t = static_cast<PlaylistModel *>(_o);
    if (_c == QMetaObject::InvokeMetaMethod) {
        switch (_id) {
        case 0: _t->imageLoaded((*reinterpret_cast<std::add_pointer_t<QImage>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<QModelIndex>>(_a[2])),(*reinterpret_cast<std::add_pointer_t<QString>>(_a[3]))); break;
        case 1: _t->onImageLoaded((*reinterpret_cast<std::add_pointer_t<QImage>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<QModelIndex>>(_a[2])),(*reinterpret_cast<std::add_pointer_t<QString>>(_a[3]))); break;
        default: ;
        }
    }
    if (_c == QMetaObject::IndexOfMethod) {
        if (QtMocHelpers::indexOfMethod<void (PlaylistModel::*)(const QImage , const QModelIndex & , const QString & )>(_a, &PlaylistModel::imageLoaded, 0))
            return;
    }
}

const QMetaObject *PlaylistModel::metaObject() const
{
    return QObject::d_ptr->metaObject ? QObject::d_ptr->dynamicMetaObject() : &staticMetaObject;
}

void *PlaylistModel::qt_metacast(const char *_clname)
{
    if (!_clname) return nullptr;
    if (!strcmp(_clname, qt_staticMetaObjectStaticContent<qt_meta_tag_ZN13PlaylistModelE_t>.strings))
        return static_cast<void*>(this);
    return QAbstractListModel::qt_metacast(_clname);
}

int PlaylistModel::qt_metacall(QMetaObject::Call _c, int _id, void **_a)
{
    _id = QAbstractListModel::qt_metacall(_c, _id, _a);
    if (_id < 0)
        return _id;
    if (_c == QMetaObject::InvokeMetaMethod) {
        if (_id < 2)
            qt_static_metacall(this, _c, _id, _a);
        _id -= 2;
    }
    if (_c == QMetaObject::RegisterMethodArgumentMetaType) {
        if (_id < 2)
            *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType();
        _id -= 2;
    }
    return _id;
}

// SIGNAL 0
void PlaylistModel::imageLoaded(const QImage _t1, const QModelIndex & _t2, const QString & _t3)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 0, nullptr, _t1, _t2, _t3);
}
namespace {
struct qt_meta_tag_ZN15ThumbnailWidgetE_t {};
} // unnamed namespace

template <> constexpr inline auto ThumbnailWidget::qt_create_metaobjectdata<qt_meta_tag_ZN15ThumbnailWidgetE_t>()
{
    namespace QMC = QtMocConstants;
    QtMocHelpers::StringRefStorage qt_stringData {
        "ThumbnailWidget",
        "filesDropped",
        "",
        "QImage",
        "image",
        "ThumbnailType",
        "type"
    };

    QtMocHelpers::UintData qt_methods {
        // Signal 'filesDropped'
        QtMocHelpers::SignalData<void(const QImage &, ThumbnailType)>(1, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 3, 4 }, { 0x80000000 | 5, 6 },
        }}),
    };
    QtMocHelpers::UintData qt_properties {
    };
    QtMocHelpers::UintData qt_enums {
    };
    return QtMocHelpers::metaObjectData<ThumbnailWidget, qt_meta_tag_ZN15ThumbnailWidgetE_t>(QMC::MetaObjectFlag{}, qt_stringData,
            qt_methods, qt_properties, qt_enums);
}
Q_CONSTINIT const QMetaObject ThumbnailWidget::staticMetaObject = { {
    QMetaObject::SuperData::link<QStackedWidget::staticMetaObject>(),
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN15ThumbnailWidgetE_t>.stringdata,
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN15ThumbnailWidgetE_t>.data,
    qt_static_metacall,
    nullptr,
    qt_staticMetaObjectRelocatingContent<qt_meta_tag_ZN15ThumbnailWidgetE_t>.metaTypes,
    nullptr
} };

void ThumbnailWidget::qt_static_metacall(QObject *_o, QMetaObject::Call _c, int _id, void **_a)
{
    auto *_t = static_cast<ThumbnailWidget *>(_o);
    if (_c == QMetaObject::InvokeMetaMethod) {
        switch (_id) {
        case 0: _t->filesDropped((*reinterpret_cast<std::add_pointer_t<QImage>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<ThumbnailType>>(_a[2]))); break;
        default: ;
        }
    }
    if (_c == QMetaObject::IndexOfMethod) {
        if (QtMocHelpers::indexOfMethod<void (ThumbnailWidget::*)(const QImage & , ThumbnailType )>(_a, &ThumbnailWidget::filesDropped, 0))
            return;
    }
}

const QMetaObject *ThumbnailWidget::metaObject() const
{
    return QObject::d_ptr->metaObject ? QObject::d_ptr->dynamicMetaObject() : &staticMetaObject;
}

void *ThumbnailWidget::qt_metacast(const char *_clname)
{
    if (!_clname) return nullptr;
    if (!strcmp(_clname, qt_staticMetaObjectStaticContent<qt_meta_tag_ZN15ThumbnailWidgetE_t>.strings))
        return static_cast<void*>(this);
    return QStackedWidget::qt_metacast(_clname);
}

int ThumbnailWidget::qt_metacall(QMetaObject::Call _c, int _id, void **_a)
{
    _id = QStackedWidget::qt_metacall(_c, _id, _a);
    if (_id < 0)
        return _id;
    if (_c == QMetaObject::InvokeMetaMethod) {
        if (_id < 1)
            qt_static_metacall(this, _c, _id, _a);
        _id -= 1;
    }
    if (_c == QMetaObject::RegisterMethodArgumentMetaType) {
        if (_id < 1)
            *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType();
        _id -= 1;
    }
    return _id;
}

// SIGNAL 0
void ThumbnailWidget::filesDropped(const QImage & _t1, ThumbnailType _t2)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 0, nullptr, _t1, _t2);
}
namespace {
struct qt_meta_tag_ZN14ThumbnailLabelE_t {};
} // unnamed namespace

template <> constexpr inline auto ThumbnailLabel::qt_create_metaobjectdata<qt_meta_tag_ZN14ThumbnailLabelE_t>()
{
    namespace QMC = QtMocConstants;
    QtMocHelpers::StringRefStorage qt_stringData {
        "ThumbnailLabel",
        "setPixmap",
        "",
        "QPixmap",
        "pixmap"
    };

    QtMocHelpers::UintData qt_methods {
        // Slot 'setPixmap'
        QtMocHelpers::SlotData<void(const QPixmap &)>(1, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 3, 4 },
        }}),
    };
    QtMocHelpers::UintData qt_properties {
    };
    QtMocHelpers::UintData qt_enums {
    };
    return QtMocHelpers::metaObjectData<ThumbnailLabel, qt_meta_tag_ZN14ThumbnailLabelE_t>(QMC::MetaObjectFlag{}, qt_stringData,
            qt_methods, qt_properties, qt_enums);
}
Q_CONSTINIT const QMetaObject ThumbnailLabel::staticMetaObject = { {
    QMetaObject::SuperData::link<QWidget::staticMetaObject>(),
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN14ThumbnailLabelE_t>.stringdata,
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN14ThumbnailLabelE_t>.data,
    qt_static_metacall,
    nullptr,
    qt_staticMetaObjectRelocatingContent<qt_meta_tag_ZN14ThumbnailLabelE_t>.metaTypes,
    nullptr
} };

void ThumbnailLabel::qt_static_metacall(QObject *_o, QMetaObject::Call _c, int _id, void **_a)
{
    auto *_t = static_cast<ThumbnailLabel *>(_o);
    if (_c == QMetaObject::InvokeMetaMethod) {
        switch (_id) {
        case 0: _t->setPixmap((*reinterpret_cast<std::add_pointer_t<QPixmap>>(_a[1]))); break;
        default: ;
        }
    }
}

const QMetaObject *ThumbnailLabel::metaObject() const
{
    return QObject::d_ptr->metaObject ? QObject::d_ptr->dynamicMetaObject() : &staticMetaObject;
}

void *ThumbnailLabel::qt_metacast(const char *_clname)
{
    if (!_clname) return nullptr;
    if (!strcmp(_clname, qt_staticMetaObjectStaticContent<qt_meta_tag_ZN14ThumbnailLabelE_t>.strings))
        return static_cast<void*>(this);
    return QWidget::qt_metacast(_clname);
}

int ThumbnailLabel::qt_metacall(QMetaObject::Call _c, int _id, void **_a)
{
    _id = QWidget::qt_metacall(_c, _id, _a);
    if (_id < 0)
        return _id;
    if (_c == QMetaObject::InvokeMetaMethod) {
        if (_id < 1)
            qt_static_metacall(this, _c, _id, _a);
        _id -= 1;
    }
    if (_c == QMetaObject::RegisterMethodArgumentMetaType) {
        if (_id < 1)
            *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType();
        _id -= 1;
    }
    return _id;
}
namespace {
struct qt_meta_tag_ZN8TreeViewE_t {};
} // unnamed namespace

template <> constexpr inline auto TreeView::qt_create_metaobjectdata<qt_meta_tag_ZN8TreeViewE_t>()
{
    namespace QMC = QtMocConstants;
    QtMocHelpers::StringRefStorage qt_stringData {
        "TreeView",
        "itemsSelected",
        "",
        "QModelIndexList",
        "selectedIndexes",
        "columnCountChanged",
        "oldCount",
        "newCount",
        "selectionChanged",
        "QItemSelection",
        "selected",
        "deselected"
    };

    QtMocHelpers::UintData qt_methods {
        // Signal 'itemsSelected'
        QtMocHelpers::SignalData<void(QModelIndexList)>(1, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 3, 4 },
        }}),
        // Slot 'columnCountChanged'
        QtMocHelpers::SlotData<void(int, int)>(5, 2, QMC::AccessProtected, QMetaType::Void, {{
            { QMetaType::Int, 6 }, { QMetaType::Int, 7 },
        }}),
        // Slot 'selectionChanged'
        QtMocHelpers::SlotData<void(const QItemSelection &, const QItemSelection &)>(8, 2, QMC::AccessProtected, QMetaType::Void, {{
            { 0x80000000 | 9, 10 }, { 0x80000000 | 9, 11 },
        }}),
    };
    QtMocHelpers::UintData qt_properties {
    };
    QtMocHelpers::UintData qt_enums {
    };
    return QtMocHelpers::metaObjectData<TreeView, qt_meta_tag_ZN8TreeViewE_t>(QMC::MetaObjectFlag{}, qt_stringData,
            qt_methods, qt_properties, qt_enums);
}
Q_CONSTINIT const QMetaObject TreeView::staticMetaObject = { {
    QMetaObject::SuperData::link<QTreeView::staticMetaObject>(),
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN8TreeViewE_t>.stringdata,
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN8TreeViewE_t>.data,
    qt_static_metacall,
    nullptr,
    qt_staticMetaObjectRelocatingContent<qt_meta_tag_ZN8TreeViewE_t>.metaTypes,
    nullptr
} };

void TreeView::qt_static_metacall(QObject *_o, QMetaObject::Call _c, int _id, void **_a)
{
    auto *_t = static_cast<TreeView *>(_o);
    if (_c == QMetaObject::InvokeMetaMethod) {
        switch (_id) {
        case 0: _t->itemsSelected((*reinterpret_cast<std::add_pointer_t<QModelIndexList>>(_a[1]))); break;
        case 1: _t->columnCountChanged((*reinterpret_cast<std::add_pointer_t<int>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<int>>(_a[2]))); break;
        case 2: _t->selectionChanged((*reinterpret_cast<std::add_pointer_t<QItemSelection>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<QItemSelection>>(_a[2]))); break;
        default: ;
        }
    }
    if (_c == QMetaObject::IndexOfMethod) {
        if (QtMocHelpers::indexOfMethod<void (TreeView::*)(QModelIndexList )>(_a, &TreeView::itemsSelected, 0))
            return;
    }
}

const QMetaObject *TreeView::metaObject() const
{
    return QObject::d_ptr->metaObject ? QObject::d_ptr->dynamicMetaObject() : &staticMetaObject;
}

void *TreeView::qt_metacast(const char *_clname)
{
    if (!_clname) return nullptr;
    if (!strcmp(_clname, qt_staticMetaObjectStaticContent<qt_meta_tag_ZN8TreeViewE_t>.strings))
        return static_cast<void*>(this);
    return QTreeView::qt_metacast(_clname);
}

int TreeView::qt_metacall(QMetaObject::Call _c, int _id, void **_a)
{
    _id = QTreeView::qt_metacall(_c, _id, _a);
    if (_id < 0)
        return _id;
    if (_c == QMetaObject::InvokeMetaMethod) {
        if (_id < 3)
            qt_static_metacall(this, _c, _id, _a);
        _id -= 3;
    }
    if (_c == QMetaObject::RegisterMethodArgumentMetaType) {
        if (_id < 3)
            *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType();
        _id -= 3;
    }
    return _id;
}

// SIGNAL 0
void TreeView::itemsSelected(QModelIndexList _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 0, nullptr, _t1);
}
namespace {
struct qt_meta_tag_ZN9TableViewE_t {};
} // unnamed namespace

template <> constexpr inline auto TableView::qt_create_metaobjectdata<qt_meta_tag_ZN9TableViewE_t>()
{
    namespace QMC = QtMocConstants;
    QtMocHelpers::StringRefStorage qt_stringData {
        "TableView"
    };

    QtMocHelpers::UintData qt_methods {
    };
    QtMocHelpers::UintData qt_properties {
    };
    QtMocHelpers::UintData qt_enums {
    };
    return QtMocHelpers::metaObjectData<TableView, qt_meta_tag_ZN9TableViewE_t>(QMC::MetaObjectFlag{}, qt_stringData,
            qt_methods, qt_properties, qt_enums);
}
Q_CONSTINIT const QMetaObject TableView::staticMetaObject = { {
    QMetaObject::SuperData::link<QTableView::staticMetaObject>(),
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN9TableViewE_t>.stringdata,
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN9TableViewE_t>.data,
    qt_static_metacall,
    nullptr,
    qt_staticMetaObjectRelocatingContent<qt_meta_tag_ZN9TableViewE_t>.metaTypes,
    nullptr
} };

void TableView::qt_static_metacall(QObject *_o, QMetaObject::Call _c, int _id, void **_a)
{
    auto *_t = static_cast<TableView *>(_o);
    (void)_t;
    (void)_c;
    (void)_id;
    (void)_a;
}

const QMetaObject *TableView::metaObject() const
{
    return QObject::d_ptr->metaObject ? QObject::d_ptr->dynamicMetaObject() : &staticMetaObject;
}

void *TableView::qt_metacast(const char *_clname)
{
    if (!_clname) return nullptr;
    if (!strcmp(_clname, qt_staticMetaObjectStaticContent<qt_meta_tag_ZN9TableViewE_t>.strings))
        return static_cast<void*>(this);
    return QTableView::qt_metacast(_clname);
}

int TableView::qt_metacall(QMetaObject::Call _c, int _id, void **_a)
{
    _id = QTableView::qt_metacall(_c, _id, _a);
    return _id;
}
namespace {
struct qt_meta_tag_ZN10ListWidgetE_t {};
} // unnamed namespace

template <> constexpr inline auto ListWidget::qt_create_metaobjectdata<qt_meta_tag_ZN10ListWidgetE_t>()
{
    namespace QMC = QtMocConstants;
    QtMocHelpers::StringRefStorage qt_stringData {
        "ListWidget",
        "enterPressed",
        "",
        "deletePressed"
    };

    QtMocHelpers::UintData qt_methods {
        // Signal 'enterPressed'
        QtMocHelpers::SignalData<void()>(1, 2, QMC::AccessPublic, QMetaType::Void),
        // Signal 'deletePressed'
        QtMocHelpers::SignalData<void()>(3, 2, QMC::AccessPublic, QMetaType::Void),
    };
    QtMocHelpers::UintData qt_properties {
    };
    QtMocHelpers::UintData qt_enums {
    };
    return QtMocHelpers::metaObjectData<ListWidget, qt_meta_tag_ZN10ListWidgetE_t>(QMC::MetaObjectFlag{}, qt_stringData,
            qt_methods, qt_properties, qt_enums);
}
Q_CONSTINIT const QMetaObject ListWidget::staticMetaObject = { {
    QMetaObject::SuperData::link<QListWidget::staticMetaObject>(),
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN10ListWidgetE_t>.stringdata,
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN10ListWidgetE_t>.data,
    qt_static_metacall,
    nullptr,
    qt_staticMetaObjectRelocatingContent<qt_meta_tag_ZN10ListWidgetE_t>.metaTypes,
    nullptr
} };

void ListWidget::qt_static_metacall(QObject *_o, QMetaObject::Call _c, int _id, void **_a)
{
    auto *_t = static_cast<ListWidget *>(_o);
    if (_c == QMetaObject::InvokeMetaMethod) {
        switch (_id) {
        case 0: _t->enterPressed(); break;
        case 1: _t->deletePressed(); break;
        default: ;
        }
    }
    if (_c == QMetaObject::IndexOfMethod) {
        if (QtMocHelpers::indexOfMethod<void (ListWidget::*)()>(_a, &ListWidget::enterPressed, 0))
            return;
        if (QtMocHelpers::indexOfMethod<void (ListWidget::*)()>(_a, &ListWidget::deletePressed, 1))
            return;
    }
}

const QMetaObject *ListWidget::metaObject() const
{
    return QObject::d_ptr->metaObject ? QObject::d_ptr->dynamicMetaObject() : &staticMetaObject;
}

void *ListWidget::qt_metacast(const char *_clname)
{
    if (!_clname) return nullptr;
    if (!strcmp(_clname, qt_staticMetaObjectStaticContent<qt_meta_tag_ZN10ListWidgetE_t>.strings))
        return static_cast<void*>(this);
    return QListWidget::qt_metacast(_clname);
}

int ListWidget::qt_metacall(QMetaObject::Call _c, int _id, void **_a)
{
    _id = QListWidget::qt_metacall(_c, _id, _a);
    if (_id < 0)
        return _id;
    if (_c == QMetaObject::InvokeMetaMethod) {
        if (_id < 2)
            qt_static_metacall(this, _c, _id, _a);
        _id -= 2;
    }
    if (_c == QMetaObject::RegisterMethodArgumentMetaType) {
        if (_id < 2)
            *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType();
        _id -= 2;
    }
    return _id;
}

// SIGNAL 0
void ListWidget::enterPressed()
{
    QMetaObject::activate(this, &staticMetaObject, 0, nullptr);
}

// SIGNAL 1
void ListWidget::deletePressed()
{
    QMetaObject::activate(this, &staticMetaObject, 1, nullptr);
}
namespace {
struct qt_meta_tag_ZN10AppHandlerE_t {};
} // unnamed namespace

template <> constexpr inline auto AppHandler::qt_create_metaobjectdata<qt_meta_tag_ZN10AppHandlerE_t>()
{
    namespace QMC = QtMocConstants;
    QtMocHelpers::StringRefStorage qt_stringData {
        "AppHandler",
        "onLastWindowClosed",
        ""
    };

    QtMocHelpers::UintData qt_methods {
        // Slot 'onLastWindowClosed'
        QtMocHelpers::SlotData<void()>(1, 2, QMC::AccessPrivate, QMetaType::Void),
    };
    QtMocHelpers::UintData qt_properties {
    };
    QtMocHelpers::UintData qt_enums {
    };
    return QtMocHelpers::metaObjectData<AppHandler, qt_meta_tag_ZN10AppHandlerE_t>(QMC::MetaObjectFlag{}, qt_stringData,
            qt_methods, qt_properties, qt_enums);
}
Q_CONSTINIT const QMetaObject AppHandler::staticMetaObject = { {
    QMetaObject::SuperData::link<QObject::staticMetaObject>(),
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN10AppHandlerE_t>.stringdata,
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN10AppHandlerE_t>.data,
    qt_static_metacall,
    nullptr,
    qt_staticMetaObjectRelocatingContent<qt_meta_tag_ZN10AppHandlerE_t>.metaTypes,
    nullptr
} };

void AppHandler::qt_static_metacall(QObject *_o, QMetaObject::Call _c, int _id, void **_a)
{
    auto *_t = static_cast<AppHandler *>(_o);
    if (_c == QMetaObject::InvokeMetaMethod) {
        switch (_id) {
        case 0: _t->onLastWindowClosed(); break;
        default: ;
        }
    }
    (void)_a;
}

const QMetaObject *AppHandler::metaObject() const
{
    return QObject::d_ptr->metaObject ? QObject::d_ptr->dynamicMetaObject() : &staticMetaObject;
}

void *AppHandler::qt_metacast(const char *_clname)
{
    if (!_clname) return nullptr;
    if (!strcmp(_clname, qt_staticMetaObjectStaticContent<qt_meta_tag_ZN10AppHandlerE_t>.strings))
        return static_cast<void*>(this);
    return QObject::qt_metacast(_clname);
}

int AppHandler::qt_metacall(QMetaObject::Call _c, int _id, void **_a)
{
    _id = QObject::qt_metacall(_c, _id, _a);
    if (_id < 0)
        return _id;
    if (_c == QMetaObject::InvokeMetaMethod) {
        if (_id < 1)
            qt_static_metacall(this, _c, _id, _a);
        _id -= 1;
    }
    if (_c == QMetaObject::RegisterMethodArgumentMetaType) {
        if (_id < 1)
            *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType();
        _id -= 1;
    }
    return _id;
}
namespace {
struct qt_meta_tag_ZN13CoreInfoLabelE_t {};
} // unnamed namespace

template <> constexpr inline auto CoreInfoLabel::qt_create_metaobjectdata<qt_meta_tag_ZN13CoreInfoLabelE_t>()
{
    namespace QMC = QtMocConstants;
    QtMocHelpers::StringRefStorage qt_stringData {
        "CoreInfoLabel"
    };

    QtMocHelpers::UintData qt_methods {
    };
    QtMocHelpers::UintData qt_properties {
    };
    QtMocHelpers::UintData qt_enums {
    };
    return QtMocHelpers::metaObjectData<CoreInfoLabel, qt_meta_tag_ZN13CoreInfoLabelE_t>(QMC::MetaObjectFlag{}, qt_stringData,
            qt_methods, qt_properties, qt_enums);
}
Q_CONSTINIT const QMetaObject CoreInfoLabel::staticMetaObject = { {
    QMetaObject::SuperData::link<QLabel::staticMetaObject>(),
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN13CoreInfoLabelE_t>.stringdata,
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN13CoreInfoLabelE_t>.data,
    qt_static_metacall,
    nullptr,
    qt_staticMetaObjectRelocatingContent<qt_meta_tag_ZN13CoreInfoLabelE_t>.metaTypes,
    nullptr
} };

void CoreInfoLabel::qt_static_metacall(QObject *_o, QMetaObject::Call _c, int _id, void **_a)
{
    auto *_t = static_cast<CoreInfoLabel *>(_o);
    (void)_t;
    (void)_c;
    (void)_id;
    (void)_a;
}

const QMetaObject *CoreInfoLabel::metaObject() const
{
    return QObject::d_ptr->metaObject ? QObject::d_ptr->dynamicMetaObject() : &staticMetaObject;
}

void *CoreInfoLabel::qt_metacast(const char *_clname)
{
    if (!_clname) return nullptr;
    if (!strcmp(_clname, qt_staticMetaObjectStaticContent<qt_meta_tag_ZN13CoreInfoLabelE_t>.strings))
        return static_cast<void*>(this);
    return QLabel::qt_metacast(_clname);
}

int CoreInfoLabel::qt_metacall(QMetaObject::Call _c, int _id, void **_a)
{
    _id = QLabel::qt_metacall(_c, _id, _a);
    return _id;
}
namespace {
struct qt_meta_tag_ZN14CoreInfoWidgetE_t {};
} // unnamed namespace

template <> constexpr inline auto CoreInfoWidget::qt_create_metaobjectdata<qt_meta_tag_ZN14CoreInfoWidgetE_t>()
{
    namespace QMC = QtMocConstants;
    QtMocHelpers::StringRefStorage qt_stringData {
        "CoreInfoWidget"
    };

    QtMocHelpers::UintData qt_methods {
    };
    QtMocHelpers::UintData qt_properties {
    };
    QtMocHelpers::UintData qt_enums {
    };
    return QtMocHelpers::metaObjectData<CoreInfoWidget, qt_meta_tag_ZN14CoreInfoWidgetE_t>(QMC::MetaObjectFlag{}, qt_stringData,
            qt_methods, qt_properties, qt_enums);
}
Q_CONSTINIT const QMetaObject CoreInfoWidget::staticMetaObject = { {
    QMetaObject::SuperData::link<QWidget::staticMetaObject>(),
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN14CoreInfoWidgetE_t>.stringdata,
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN14CoreInfoWidgetE_t>.data,
    qt_static_metacall,
    nullptr,
    qt_staticMetaObjectRelocatingContent<qt_meta_tag_ZN14CoreInfoWidgetE_t>.metaTypes,
    nullptr
} };

void CoreInfoWidget::qt_static_metacall(QObject *_o, QMetaObject::Call _c, int _id, void **_a)
{
    auto *_t = static_cast<CoreInfoWidget *>(_o);
    (void)_t;
    (void)_c;
    (void)_id;
    (void)_a;
}

const QMetaObject *CoreInfoWidget::metaObject() const
{
    return QObject::d_ptr->metaObject ? QObject::d_ptr->dynamicMetaObject() : &staticMetaObject;
}

void *CoreInfoWidget::qt_metacast(const char *_clname)
{
    if (!_clname) return nullptr;
    if (!strcmp(_clname, qt_staticMetaObjectStaticContent<qt_meta_tag_ZN14CoreInfoWidgetE_t>.strings))
        return static_cast<void*>(this);
    return QWidget::qt_metacast(_clname);
}

int CoreInfoWidget::qt_metacall(QMetaObject::Call _c, int _id, void **_a)
{
    _id = QWidget::qt_metacall(_c, _id, _a);
    return _id;
}
namespace {
struct qt_meta_tag_ZN11LogTextEditE_t {};
} // unnamed namespace

template <> constexpr inline auto LogTextEdit::qt_create_metaobjectdata<qt_meta_tag_ZN11LogTextEditE_t>()
{
    namespace QMC = QtMocConstants;
    QtMocHelpers::StringRefStorage qt_stringData {
        "LogTextEdit",
        "appendMessage",
        "",
        "text"
    };

    QtMocHelpers::UintData qt_methods {
        // Slot 'appendMessage'
        QtMocHelpers::SlotData<void(const QString &)>(1, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::QString, 3 },
        }}),
    };
    QtMocHelpers::UintData qt_properties {
    };
    QtMocHelpers::UintData qt_enums {
    };
    return QtMocHelpers::metaObjectData<LogTextEdit, qt_meta_tag_ZN11LogTextEditE_t>(QMC::MetaObjectFlag{}, qt_stringData,
            qt_methods, qt_properties, qt_enums);
}
Q_CONSTINIT const QMetaObject LogTextEdit::staticMetaObject = { {
    QMetaObject::SuperData::link<QPlainTextEdit::staticMetaObject>(),
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN11LogTextEditE_t>.stringdata,
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN11LogTextEditE_t>.data,
    qt_static_metacall,
    nullptr,
    qt_staticMetaObjectRelocatingContent<qt_meta_tag_ZN11LogTextEditE_t>.metaTypes,
    nullptr
} };

void LogTextEdit::qt_static_metacall(QObject *_o, QMetaObject::Call _c, int _id, void **_a)
{
    auto *_t = static_cast<LogTextEdit *>(_o);
    if (_c == QMetaObject::InvokeMetaMethod) {
        switch (_id) {
        case 0: _t->appendMessage((*reinterpret_cast<std::add_pointer_t<QString>>(_a[1]))); break;
        default: ;
        }
    }
}

const QMetaObject *LogTextEdit::metaObject() const
{
    return QObject::d_ptr->metaObject ? QObject::d_ptr->dynamicMetaObject() : &staticMetaObject;
}

void *LogTextEdit::qt_metacast(const char *_clname)
{
    if (!_clname) return nullptr;
    if (!strcmp(_clname, qt_staticMetaObjectStaticContent<qt_meta_tag_ZN11LogTextEditE_t>.strings))
        return static_cast<void*>(this);
    return QPlainTextEdit::qt_metacast(_clname);
}

int LogTextEdit::qt_metacall(QMetaObject::Call _c, int _id, void **_a)
{
    _id = QPlainTextEdit::qt_metacall(_c, _id, _a);
    if (_id < 0)
        return _id;
    if (_c == QMetaObject::InvokeMetaMethod) {
        if (_id < 1)
            qt_static_metacall(this, _c, _id, _a);
        _id -= 1;
    }
    if (_c == QMetaObject::RegisterMethodArgumentMetaType) {
        if (_id < 1)
            *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType();
        _id -= 1;
    }
    return _id;
}
namespace {
struct qt_meta_tag_ZN8GridItemE_t {};
} // unnamed namespace

template <> constexpr inline auto GridItem::qt_create_metaobjectdata<qt_meta_tag_ZN8GridItemE_t>()
{
    namespace QMC = QtMocConstants;
    QtMocHelpers::StringRefStorage qt_stringData {
        "GridItem",
        "thumbnailvalign",
        "padding"
    };

    QtMocHelpers::UintData qt_methods {
    };
    QtMocHelpers::UintData qt_properties {
        // property 'thumbnailvalign'
        QtMocHelpers::PropertyData<QString>(1, QMetaType::QString, QMC::DefaultPropertyFlags | QMC::Writable),
        // property 'padding'
        QtMocHelpers::PropertyData<int>(2, QMetaType::Int, QMC::DefaultPropertyFlags | QMC::Writable | QMC::StdCppSet),
    };
    QtMocHelpers::UintData qt_enums {
    };
    return QtMocHelpers::metaObjectData<GridItem, qt_meta_tag_ZN8GridItemE_t>(QMC::MetaObjectFlag{}, qt_stringData,
            qt_methods, qt_properties, qt_enums);
}
Q_CONSTINIT const QMetaObject GridItem::staticMetaObject = { {
    QMetaObject::SuperData::link<QWidget::staticMetaObject>(),
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN8GridItemE_t>.stringdata,
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN8GridItemE_t>.data,
    qt_static_metacall,
    nullptr,
    qt_staticMetaObjectRelocatingContent<qt_meta_tag_ZN8GridItemE_t>.metaTypes,
    nullptr
} };

void GridItem::qt_static_metacall(QObject *_o, QMetaObject::Call _c, int _id, void **_a)
{
    auto *_t = static_cast<GridItem *>(_o);
    if (_c == QMetaObject::ReadProperty) {
        void *_v = _a[0];
        switch (_id) {
        case 0: *reinterpret_cast<QString*>(_v) = _t->getThumbnailVerticalAlign(); break;
        case 1: *reinterpret_cast<int*>(_v) = _t->getPadding(); break;
        default: break;
        }
    }
    if (_c == QMetaObject::WriteProperty) {
        void *_v = _a[0];
        switch (_id) {
        case 0: _t->setThumbnailVerticalAlign(*reinterpret_cast<QString*>(_v)); break;
        case 1: _t->setPadding(*reinterpret_cast<int*>(_v)); break;
        default: break;
        }
    }
}

const QMetaObject *GridItem::metaObject() const
{
    return QObject::d_ptr->metaObject ? QObject::d_ptr->dynamicMetaObject() : &staticMetaObject;
}

void *GridItem::qt_metacast(const char *_clname)
{
    if (!_clname) return nullptr;
    if (!strcmp(_clname, qt_staticMetaObjectStaticContent<qt_meta_tag_ZN8GridItemE_t>.strings))
        return static_cast<void*>(this);
    return QWidget::qt_metacast(_clname);
}

int GridItem::qt_metacall(QMetaObject::Call _c, int _id, void **_a)
{
    _id = QWidget::qt_metacall(_c, _id, _a);
    if (_id < 0)
        return _id;
    if (_c == QMetaObject::ReadProperty || _c == QMetaObject::WriteProperty
            || _c == QMetaObject::ResetProperty || _c == QMetaObject::BindableProperty
            || _c == QMetaObject::RegisterPropertyMetaType) {
        qt_static_metacall(this, _c, _id, _a);
        _id -= 2;
    }
    return _id;
}
namespace {
struct qt_meta_tag_ZN10MainWindowE_t {};
} // unnamed namespace

template <> constexpr inline auto MainWindow::qt_create_metaobjectdata<qt_meta_tag_ZN10MainWindowE_t>()
{
    namespace QMC = QtMocConstants;
    QtMocHelpers::StringRefStorage qt_stringData {
        "MainWindow",
        "thumbnailChanged",
        "",
        "QPixmap",
        "pixmap",
        "thumbnail2Changed",
        "thumbnail3Changed",
        "thumbnail4Changed",
        "gotLogMessage",
        "msg",
        "gotStatusMessage",
        "priority",
        "duration",
        "flush",
        "gotReloadPlaylists",
        "gotReloadShaderParams",
        "gotReloadCoreOptions",
        "showErrorMessageDeferred",
        "showInfoMessageDeferred",
        "extractArchiveDeferred",
        "path",
        "extractionDir",
        "tempExtension",
        "retro_task_callback_t",
        "cb",
        "itemChanged",
        "updateThumbnails",
        "gridItemChanged",
        "title",
        "gotThumbnailDownload",
        "system",
        "scrollToDownloads",
        "scrollToDownloadsAgain",
        "onBrowserDownloadsClicked",
        "onBrowserUpClicked",
        "onBrowserStartClicked",
        "initContentTableWidget",
        "onViewClosedDocksAboutToShow",
        "onShowHiddenDockWidgetAction",
        "setCoreActions",
        "onRunClicked",
        "loadContent",
        "QHash<QString,QString>",
        "contentHash",
        "onStartCoreClicked",
        "onDropWidgetEnterPressed",
        "selectBrowserDir",
        "setThumbnail",
        "widgetName",
        "QPixmap&",
        "acceptDrop",
        "onResizeThumbnailOne",
        "onResizeThumbnailTwo",
        "onResizeThumbnailThree",
        "onResizeThumbnailFour",
        "appendLogMessage",
        "onGotLogMessage",
        "onGotStatusMessage",
        "reloadPlaylists",
        "deferReloadPlaylists",
        "onGotReloadPlaylists",
        "onGotReloadShaderParams",
        "onGotReloadCoreOptions",
        "showWelcomeScreen",
        "onIconViewClicked",
        "onListViewClicked",
        "onBoxartThumbnailClicked",
        "onScreenshotThumbnailClicked",
        "onTitleThumbnailClicked",
        "onLogoThumbnailClicked",
        "onTabWidgetIndexChanged",
        "index",
        "deleteCurrentPlaylistItem",
        "onFileDropWidgetContextMenuRequested",
        "QPoint",
        "pos",
        "showAbout",
        "showDocs",
        "onThumbnailPackExtractFinished",
        "success",
        "deferReloadShaderParams",
        "downloadThumbnail",
        "QUrl",
        "url",
        "downloadAllThumbnails",
        "downloadPlaylistThumbnails",
        "playlistPath",
        "downloadNextPlaylistThumbnail",
        "type",
        "changeThumbnailType",
        "ThumbnailType",
        "onThumbnailDropped",
        "QImage",
        "image",
        "onLoadCoreClicked",
        "extensionFilters",
        "onUnloadCoreMenuAction",
        "onTimeout",
        "onCoreLoaded",
        "onCurrentTableItemDataChanged",
        "QModelIndex",
        "topLeft",
        "bottomRight",
        "QList<int>",
        "roles",
        "onCurrentListItemChanged",
        "QListWidgetItem*",
        "current",
        "previous",
        "onCurrentListItemDataChanged",
        "item",
        "onCurrentItemChanged",
        "hash",
        "onCurrentFileChanged",
        "onSearchEnterPressed",
        "onSearchLineEditEdited",
        "text",
        "onContentItemDoubleClicked",
        "onFileDoubleClicked",
        "onCoreLoadWindowClosed",
        "onTreeViewItemsSelected",
        "QModelIndexList",
        "selectedIndexes",
        "onSearchResetClicked",
        "onLaunchWithComboBoxIndexChanged",
        "onFileBrowserTreeContextMenuRequested",
        "onPlaylistWidgetContextMenuRequested",
        "onStopClicked",
        "onZoomValueChanged",
        "value",
        "onPlaylistFilesDropped",
        "files",
        "onShaderParamsClicked",
        "onCoreOptionsClicked",
        "onShowErrorMessage",
        "onShowInfoMessage",
        "onContributorsClicked",
        "onItemChanged",
        "onFileSystemDirLoaded",
        "onFileBrowserTableDirLoaded",
        "onDownloadScroll",
        "onDownloadScrollAgain",
        "onExtractArchive",
        "onThumbnailDownloadNetworkError",
        "QNetworkReply::NetworkError",
        "code",
        "onThumbnailDownloadNetworkSslErrors",
        "QList<QSslError>",
        "errors",
        "onThumbnailDownloadFinished",
        "onThumbnailDownloadProgress",
        "bytesReceived",
        "bytesTotal",
        "onThumbnailDownloadReadyRead",
        "onThumbnailDownloadCanceled",
        "onDownloadThumbnail",
        "onThumbnailPackDownloadNetworkError",
        "onThumbnailPackDownloadNetworkSslErrors",
        "onThumbnailPackDownloadFinished",
        "onThumbnailPackDownloadProgress",
        "onThumbnailPackDownloadReadyRead",
        "onThumbnailPackDownloadCanceled",
        "onPlaylistThumbnailDownloadNetworkError",
        "onPlaylistThumbnailDownloadNetworkSslErrors",
        "onPlaylistThumbnailDownloadFinished",
        "onPlaylistThumbnailDownloadProgress",
        "onPlaylistThumbnailDownloadReadyRead",
        "onPlaylistThumbnailDownloadCanceled",
        "startTimer",
        "updateVisibleItems"
    };

    QtMocHelpers::UintData qt_methods {
        // Signal 'thumbnailChanged'
        QtMocHelpers::SignalData<void(const QPixmap &)>(1, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 3, 4 },
        }}),
        // Signal 'thumbnail2Changed'
        QtMocHelpers::SignalData<void(const QPixmap &)>(5, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 3, 4 },
        }}),
        // Signal 'thumbnail3Changed'
        QtMocHelpers::SignalData<void(const QPixmap &)>(6, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 3, 4 },
        }}),
        // Signal 'thumbnail4Changed'
        QtMocHelpers::SignalData<void(const QPixmap &)>(7, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 3, 4 },
        }}),
        // Signal 'gotLogMessage'
        QtMocHelpers::SignalData<void(const QString &)>(8, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::QString, 9 },
        }}),
        // Signal 'gotStatusMessage'
        QtMocHelpers::SignalData<void(QString, unsigned , unsigned , bool)>(10, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::QString, 9 }, { QMetaType::UInt, 11 }, { QMetaType::UInt, 12 }, { QMetaType::Bool, 13 },
        }}),
        // Signal 'gotReloadPlaylists'
        QtMocHelpers::SignalData<void()>(14, 2, QMC::AccessPublic, QMetaType::Void),
        // Signal 'gotReloadShaderParams'
        QtMocHelpers::SignalData<void()>(15, 2, QMC::AccessPublic, QMetaType::Void),
        // Signal 'gotReloadCoreOptions'
        QtMocHelpers::SignalData<void()>(16, 2, QMC::AccessPublic, QMetaType::Void),
        // Signal 'showErrorMessageDeferred'
        QtMocHelpers::SignalData<void(QString)>(17, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::QString, 9 },
        }}),
        // Signal 'showInfoMessageDeferred'
        QtMocHelpers::SignalData<void(QString)>(18, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::QString, 9 },
        }}),
        // Signal 'extractArchiveDeferred'
        QtMocHelpers::SignalData<void(QString, QString, QString, retro_task_callback_t)>(19, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::QString, 20 }, { QMetaType::QString, 21 }, { QMetaType::QString, 22 }, { 0x80000000 | 23, 24 },
        }}),
        // Signal 'itemChanged'
        QtMocHelpers::SignalData<void()>(25, 2, QMC::AccessPublic, QMetaType::Void),
        // Signal 'updateThumbnails'
        QtMocHelpers::SignalData<void()>(26, 2, QMC::AccessPublic, QMetaType::Void),
        // Signal 'gridItemChanged'
        QtMocHelpers::SignalData<void(QString)>(27, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::QString, 28 },
        }}),
        // Signal 'gotThumbnailDownload'
        QtMocHelpers::SignalData<void(QString, QString)>(29, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::QString, 30 }, { QMetaType::QString, 28 },
        }}),
        // Signal 'scrollToDownloads'
        QtMocHelpers::SignalData<void(QString)>(31, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::QString, 20 },
        }}),
        // Signal 'scrollToDownloadsAgain'
        QtMocHelpers::SignalData<void(QString)>(32, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::QString, 20 },
        }}),
        // Slot 'onBrowserDownloadsClicked'
        QtMocHelpers::SlotData<void()>(33, 2, QMC::AccessPublic, QMetaType::Void),
        // Slot 'onBrowserUpClicked'
        QtMocHelpers::SlotData<void()>(34, 2, QMC::AccessPublic, QMetaType::Void),
        // Slot 'onBrowserStartClicked'
        QtMocHelpers::SlotData<void()>(35, 2, QMC::AccessPublic, QMetaType::Void),
        // Slot 'initContentTableWidget'
        QtMocHelpers::SlotData<void()>(36, 2, QMC::AccessPublic, QMetaType::Void),
        // Slot 'onViewClosedDocksAboutToShow'
        QtMocHelpers::SlotData<void()>(37, 2, QMC::AccessPublic, QMetaType::Void),
        // Slot 'onShowHiddenDockWidgetAction'
        QtMocHelpers::SlotData<void()>(38, 2, QMC::AccessPublic, QMetaType::Void),
        // Slot 'setCoreActions'
        QtMocHelpers::SlotData<void()>(39, 2, QMC::AccessPublic, QMetaType::Void),
        // Slot 'onRunClicked'
        QtMocHelpers::SlotData<void()>(40, 2, QMC::AccessPublic, QMetaType::Void),
        // Slot 'loadContent'
        QtMocHelpers::SlotData<void(const QHash<QString,QString> &)>(41, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 42, 43 },
        }}),
        // Slot 'onStartCoreClicked'
        QtMocHelpers::SlotData<void()>(44, 2, QMC::AccessPublic, QMetaType::Void),
        // Slot 'onDropWidgetEnterPressed'
        QtMocHelpers::SlotData<void()>(45, 2, QMC::AccessPublic, QMetaType::Void),
        // Slot 'selectBrowserDir'
        QtMocHelpers::SlotData<void(QString)>(46, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::QString, 20 },
        }}),
        // Slot 'setThumbnail'
        QtMocHelpers::SlotData<void(QString, QPixmap &, bool)>(47, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::QString, 48 }, { 0x80000000 | 49, 4 }, { QMetaType::Bool, 50 },
        }}),
        // Slot 'onResizeThumbnailOne'
        QtMocHelpers::SlotData<void(QPixmap &, bool)>(51, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 49, 4 }, { QMetaType::Bool, 50 },
        }}),
        // Slot 'onResizeThumbnailTwo'
        QtMocHelpers::SlotData<void(QPixmap &, bool)>(52, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 49, 4 }, { QMetaType::Bool, 50 },
        }}),
        // Slot 'onResizeThumbnailThree'
        QtMocHelpers::SlotData<void(QPixmap &, bool)>(53, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 49, 4 }, { QMetaType::Bool, 50 },
        }}),
        // Slot 'onResizeThumbnailFour'
        QtMocHelpers::SlotData<void(QPixmap &, bool)>(54, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 49, 4 }, { QMetaType::Bool, 50 },
        }}),
        // Slot 'appendLogMessage'
        QtMocHelpers::SlotData<void(const QString &)>(55, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::QString, 9 },
        }}),
        // Slot 'onGotLogMessage'
        QtMocHelpers::SlotData<void(const QString &)>(56, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::QString, 9 },
        }}),
        // Slot 'onGotStatusMessage'
        QtMocHelpers::SlotData<void(QString, unsigned , unsigned , bool)>(57, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::QString, 9 }, { QMetaType::UInt, 11 }, { QMetaType::UInt, 12 }, { QMetaType::Bool, 13 },
        }}),
        // Slot 'reloadPlaylists'
        QtMocHelpers::SlotData<void()>(58, 2, QMC::AccessPublic, QMetaType::Void),
        // Slot 'deferReloadPlaylists'
        QtMocHelpers::SlotData<void()>(59, 2, QMC::AccessPublic, QMetaType::Void),
        // Slot 'onGotReloadPlaylists'
        QtMocHelpers::SlotData<void()>(60, 2, QMC::AccessPublic, QMetaType::Void),
        // Slot 'onGotReloadShaderParams'
        QtMocHelpers::SlotData<void()>(61, 2, QMC::AccessPublic, QMetaType::Void),
        // Slot 'onGotReloadCoreOptions'
        QtMocHelpers::SlotData<void()>(62, 2, QMC::AccessPublic, QMetaType::Void),
        // Slot 'showWelcomeScreen'
        QtMocHelpers::SlotData<void()>(63, 2, QMC::AccessPublic, QMetaType::Void),
        // Slot 'onIconViewClicked'
        QtMocHelpers::SlotData<void()>(64, 2, QMC::AccessPublic, QMetaType::Void),
        // Slot 'onListViewClicked'
        QtMocHelpers::SlotData<void()>(65, 2, QMC::AccessPublic, QMetaType::Void),
        // Slot 'onBoxartThumbnailClicked'
        QtMocHelpers::SlotData<void()>(66, 2, QMC::AccessPublic, QMetaType::Void),
        // Slot 'onScreenshotThumbnailClicked'
        QtMocHelpers::SlotData<void()>(67, 2, QMC::AccessPublic, QMetaType::Void),
        // Slot 'onTitleThumbnailClicked'
        QtMocHelpers::SlotData<void()>(68, 2, QMC::AccessPublic, QMetaType::Void),
        // Slot 'onLogoThumbnailClicked'
        QtMocHelpers::SlotData<void()>(69, 2, QMC::AccessPublic, QMetaType::Void),
        // Slot 'onTabWidgetIndexChanged'
        QtMocHelpers::SlotData<void(int)>(70, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::Int, 71 },
        }}),
        // Slot 'deleteCurrentPlaylistItem'
        QtMocHelpers::SlotData<void()>(72, 2, QMC::AccessPublic, QMetaType::Void),
        // Slot 'onFileDropWidgetContextMenuRequested'
        QtMocHelpers::SlotData<void(const QPoint &)>(73, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 74, 75 },
        }}),
        // Slot 'showAbout'
        QtMocHelpers::SlotData<void()>(76, 2, QMC::AccessPublic, QMetaType::Void),
        // Slot 'showDocs'
        QtMocHelpers::SlotData<void()>(77, 2, QMC::AccessPublic, QMetaType::Void),
        // Slot 'onThumbnailPackExtractFinished'
        QtMocHelpers::SlotData<void(bool)>(78, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::Bool, 79 },
        }}),
        // Slot 'deferReloadShaderParams'
        QtMocHelpers::SlotData<void()>(80, 2, QMC::AccessPublic, QMetaType::Void),
        // Slot 'downloadThumbnail'
        QtMocHelpers::SlotData<void(QString, QString, QUrl)>(81, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::QString, 30 }, { QMetaType::QString, 28 }, { 0x80000000 | 82, 83 },
        }}),
        // Slot 'downloadThumbnail'
        QtMocHelpers::SlotData<void(QString, QString)>(81, 2, QMC::AccessPublic | QMC::MethodCloned, QMetaType::Void, {{
            { QMetaType::QString, 30 }, { QMetaType::QString, 28 },
        }}),
        // Slot 'downloadAllThumbnails'
        QtMocHelpers::SlotData<void(QString, QUrl)>(84, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::QString, 30 }, { 0x80000000 | 82, 83 },
        }}),
        // Slot 'downloadAllThumbnails'
        QtMocHelpers::SlotData<void(QString)>(84, 2, QMC::AccessPublic | QMC::MethodCloned, QMetaType::Void, {{
            { QMetaType::QString, 30 },
        }}),
        // Slot 'downloadPlaylistThumbnails'
        QtMocHelpers::SlotData<void(QString)>(85, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::QString, 86 },
        }}),
        // Slot 'downloadNextPlaylistThumbnail'
        QtMocHelpers::SlotData<void(QString, QString, QString, QUrl)>(87, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::QString, 30 }, { QMetaType::QString, 28 }, { QMetaType::QString, 88 }, { 0x80000000 | 82, 83 },
        }}),
        // Slot 'downloadNextPlaylistThumbnail'
        QtMocHelpers::SlotData<void(QString, QString, QString)>(87, 2, QMC::AccessPublic | QMC::MethodCloned, QMetaType::Void, {{
            { QMetaType::QString, 30 }, { QMetaType::QString, 28 }, { QMetaType::QString, 88 },
        }}),
        // Slot 'changeThumbnailType'
        QtMocHelpers::SlotData<void(ThumbnailType)>(89, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 90, 88 },
        }}),
        // Slot 'onThumbnailDropped'
        QtMocHelpers::SlotData<void(const QImage &, ThumbnailType)>(91, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 92, 93 }, { 0x80000000 | 90, 88 },
        }}),
        // Slot 'onLoadCoreClicked'
        QtMocHelpers::SlotData<void(const QStringList &)>(94, 2, QMC::AccessPrivate, QMetaType::Void, {{
            { QMetaType::QStringList, 95 },
        }}),
        // Slot 'onLoadCoreClicked'
        QtMocHelpers::SlotData<void()>(94, 2, QMC::AccessPrivate | QMC::MethodCloned, QMetaType::Void),
        // Slot 'onUnloadCoreMenuAction'
        QtMocHelpers::SlotData<void()>(96, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'onTimeout'
        QtMocHelpers::SlotData<void()>(97, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'onCoreLoaded'
        QtMocHelpers::SlotData<void()>(98, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'onCurrentTableItemDataChanged'
        QtMocHelpers::SlotData<void(const QModelIndex &, const QModelIndex &, const QVector<int> &)>(99, 2, QMC::AccessPrivate, QMetaType::Void, {{
            { 0x80000000 | 100, 101 }, { 0x80000000 | 100, 102 }, { 0x80000000 | 103, 104 },
        }}),
        // Slot 'onCurrentListItemChanged'
        QtMocHelpers::SlotData<void(QListWidgetItem *, QListWidgetItem *)>(105, 2, QMC::AccessPrivate, QMetaType::Void, {{
            { 0x80000000 | 106, 107 }, { 0x80000000 | 106, 108 },
        }}),
        // Slot 'onCurrentListItemDataChanged'
        QtMocHelpers::SlotData<void(QListWidgetItem *)>(109, 2, QMC::AccessPrivate, QMetaType::Void, {{
            { 0x80000000 | 106, 110 },
        }}),
        // Slot 'onCurrentItemChanged'
        QtMocHelpers::SlotData<void(const QModelIndex &)>(111, 2, QMC::AccessPrivate, QMetaType::Void, {{
            { 0x80000000 | 100, 71 },
        }}),
        // Slot 'onCurrentItemChanged'
        QtMocHelpers::SlotData<void(const QHash<QString,QString> &)>(111, 2, QMC::AccessPrivate, QMetaType::Void, {{
            { 0x80000000 | 42, 112 },
        }}),
        // Slot 'onCurrentFileChanged'
        QtMocHelpers::SlotData<void(const QModelIndex &)>(113, 2, QMC::AccessPrivate, QMetaType::Void, {{
            { 0x80000000 | 100, 71 },
        }}),
        // Slot 'onSearchEnterPressed'
        QtMocHelpers::SlotData<void()>(114, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'onSearchLineEditEdited'
        QtMocHelpers::SlotData<void(const QString &)>(115, 2, QMC::AccessPrivate, QMetaType::Void, {{
            { QMetaType::QString, 116 },
        }}),
        // Slot 'onContentItemDoubleClicked'
        QtMocHelpers::SlotData<void(const QModelIndex &)>(117, 2, QMC::AccessPrivate, QMetaType::Void, {{
            { 0x80000000 | 100, 71 },
        }}),
        // Slot 'onFileDoubleClicked'
        QtMocHelpers::SlotData<void(const QModelIndex &)>(118, 2, QMC::AccessPrivate, QMetaType::Void, {{
            { 0x80000000 | 100, 71 },
        }}),
        // Slot 'onCoreLoadWindowClosed'
        QtMocHelpers::SlotData<void()>(119, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'onTreeViewItemsSelected'
        QtMocHelpers::SlotData<void(QModelIndexList)>(120, 2, QMC::AccessPrivate, QMetaType::Void, {{
            { 0x80000000 | 121, 122 },
        }}),
        // Slot 'onSearchResetClicked'
        QtMocHelpers::SlotData<void()>(123, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'onLaunchWithComboBoxIndexChanged'
        QtMocHelpers::SlotData<void(int)>(124, 2, QMC::AccessPrivate, QMetaType::Void, {{
            { QMetaType::Int, 71 },
        }}),
        // Slot 'onFileBrowserTreeContextMenuRequested'
        QtMocHelpers::SlotData<void(const QPoint &)>(125, 2, QMC::AccessPrivate, QMetaType::Void, {{
            { 0x80000000 | 74, 75 },
        }}),
        // Slot 'onPlaylistWidgetContextMenuRequested'
        QtMocHelpers::SlotData<void(const QPoint &)>(126, 2, QMC::AccessPrivate, QMetaType::Void, {{
            { 0x80000000 | 74, 75 },
        }}),
        // Slot 'onStopClicked'
        QtMocHelpers::SlotData<void()>(127, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'onZoomValueChanged'
        QtMocHelpers::SlotData<void(int)>(128, 2, QMC::AccessPrivate, QMetaType::Void, {{
            { QMetaType::Int, 129 },
        }}),
        // Slot 'onPlaylistFilesDropped'
        QtMocHelpers::SlotData<void(QStringList)>(130, 2, QMC::AccessPrivate, QMetaType::Void, {{
            { QMetaType::QStringList, 131 },
        }}),
        // Slot 'onShaderParamsClicked'
        QtMocHelpers::SlotData<void()>(132, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'onCoreOptionsClicked'
        QtMocHelpers::SlotData<void()>(133, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'onShowErrorMessage'
        QtMocHelpers::SlotData<void(QString)>(134, 2, QMC::AccessPrivate, QMetaType::Void, {{
            { QMetaType::QString, 9 },
        }}),
        // Slot 'onShowInfoMessage'
        QtMocHelpers::SlotData<void(QString)>(135, 2, QMC::AccessPrivate, QMetaType::Void, {{
            { QMetaType::QString, 9 },
        }}),
        // Slot 'onContributorsClicked'
        QtMocHelpers::SlotData<void()>(136, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'onItemChanged'
        QtMocHelpers::SlotData<void()>(137, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'onFileSystemDirLoaded'
        QtMocHelpers::SlotData<void(const QString &)>(138, 2, QMC::AccessPrivate, QMetaType::Void, {{
            { QMetaType::QString, 20 },
        }}),
        // Slot 'onFileBrowserTableDirLoaded'
        QtMocHelpers::SlotData<void(const QString &)>(139, 2, QMC::AccessPrivate, QMetaType::Void, {{
            { QMetaType::QString, 20 },
        }}),
        // Slot 'onDownloadScroll'
        QtMocHelpers::SlotData<void(QString)>(140, 2, QMC::AccessPrivate, QMetaType::Void, {{
            { QMetaType::QString, 20 },
        }}),
        // Slot 'onDownloadScrollAgain'
        QtMocHelpers::SlotData<void(QString)>(141, 2, QMC::AccessPrivate, QMetaType::Void, {{
            { QMetaType::QString, 20 },
        }}),
        // Slot 'onExtractArchive'
        QtMocHelpers::SlotData<int(QString, QString, QString, retro_task_callback_t)>(142, 2, QMC::AccessPrivate, QMetaType::Int, {{
            { QMetaType::QString, 20 }, { QMetaType::QString, 21 }, { QMetaType::QString, 22 }, { 0x80000000 | 23, 24 },
        }}),
        // Slot 'onThumbnailDownloadNetworkError'
        QtMocHelpers::SlotData<void(QNetworkReply::NetworkError)>(143, 2, QMC::AccessPrivate, QMetaType::Void, {{
            { 0x80000000 | 144, 145 },
        }}),
        // Slot 'onThumbnailDownloadNetworkSslErrors'
        QtMocHelpers::SlotData<void(const QList<QSslError> &)>(146, 2, QMC::AccessPrivate, QMetaType::Void, {{
            { 0x80000000 | 147, 148 },
        }}),
        // Slot 'onThumbnailDownloadFinished'
        QtMocHelpers::SlotData<void()>(149, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'onThumbnailDownloadProgress'
        QtMocHelpers::SlotData<void(qint64, qint64)>(150, 2, QMC::AccessPrivate, QMetaType::Void, {{
            { QMetaType::LongLong, 151 }, { QMetaType::LongLong, 152 },
        }}),
        // Slot 'onThumbnailDownloadReadyRead'
        QtMocHelpers::SlotData<void()>(153, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'onThumbnailDownloadCanceled'
        QtMocHelpers::SlotData<void()>(154, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'onDownloadThumbnail'
        QtMocHelpers::SlotData<void(QString, QString)>(155, 2, QMC::AccessPrivate, QMetaType::Void, {{
            { QMetaType::QString, 30 }, { QMetaType::QString, 28 },
        }}),
        // Slot 'onThumbnailPackDownloadNetworkError'
        QtMocHelpers::SlotData<void(QNetworkReply::NetworkError)>(156, 2, QMC::AccessPrivate, QMetaType::Void, {{
            { 0x80000000 | 144, 145 },
        }}),
        // Slot 'onThumbnailPackDownloadNetworkSslErrors'
        QtMocHelpers::SlotData<void(const QList<QSslError> &)>(157, 2, QMC::AccessPrivate, QMetaType::Void, {{
            { 0x80000000 | 147, 148 },
        }}),
        // Slot 'onThumbnailPackDownloadFinished'
        QtMocHelpers::SlotData<void()>(158, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'onThumbnailPackDownloadProgress'
        QtMocHelpers::SlotData<void(qint64, qint64)>(159, 2, QMC::AccessPrivate, QMetaType::Void, {{
            { QMetaType::LongLong, 151 }, { QMetaType::LongLong, 152 },
        }}),
        // Slot 'onThumbnailPackDownloadReadyRead'
        QtMocHelpers::SlotData<void()>(160, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'onThumbnailPackDownloadCanceled'
        QtMocHelpers::SlotData<void()>(161, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'onPlaylistThumbnailDownloadNetworkError'
        QtMocHelpers::SlotData<void(QNetworkReply::NetworkError)>(162, 2, QMC::AccessPrivate, QMetaType::Void, {{
            { 0x80000000 | 144, 145 },
        }}),
        // Slot 'onPlaylistThumbnailDownloadNetworkSslErrors'
        QtMocHelpers::SlotData<void(const QList<QSslError> &)>(163, 2, QMC::AccessPrivate, QMetaType::Void, {{
            { 0x80000000 | 147, 148 },
        }}),
        // Slot 'onPlaylistThumbnailDownloadFinished'
        QtMocHelpers::SlotData<void()>(164, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'onPlaylistThumbnailDownloadProgress'
        QtMocHelpers::SlotData<void(qint64, qint64)>(165, 2, QMC::AccessPrivate, QMetaType::Void, {{
            { QMetaType::LongLong, 151 }, { QMetaType::LongLong, 152 },
        }}),
        // Slot 'onPlaylistThumbnailDownloadReadyRead'
        QtMocHelpers::SlotData<void()>(166, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'onPlaylistThumbnailDownloadCanceled'
        QtMocHelpers::SlotData<void()>(167, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'startTimer'
        QtMocHelpers::SlotData<void()>(168, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'updateVisibleItems'
        QtMocHelpers::SlotData<void()>(169, 2, QMC::AccessPrivate, QMetaType::Void),
    };
    QtMocHelpers::UintData qt_properties {
    };
    QtMocHelpers::UintData qt_enums {
    };
    return QtMocHelpers::metaObjectData<MainWindow, qt_meta_tag_ZN10MainWindowE_t>(QMC::MetaObjectFlag{}, qt_stringData,
            qt_methods, qt_properties, qt_enums);
}
Q_CONSTINIT const QMetaObject MainWindow::staticMetaObject = { {
    QMetaObject::SuperData::link<QMainWindow::staticMetaObject>(),
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN10MainWindowE_t>.stringdata,
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN10MainWindowE_t>.data,
    qt_static_metacall,
    nullptr,
    qt_staticMetaObjectRelocatingContent<qt_meta_tag_ZN10MainWindowE_t>.metaTypes,
    nullptr
} };

void MainWindow::qt_static_metacall(QObject *_o, QMetaObject::Call _c, int _id, void **_a)
{
    auto *_t = static_cast<MainWindow *>(_o);
    if (_c == QMetaObject::InvokeMetaMethod) {
        switch (_id) {
        case 0: _t->thumbnailChanged((*reinterpret_cast<std::add_pointer_t<QPixmap>>(_a[1]))); break;
        case 1: _t->thumbnail2Changed((*reinterpret_cast<std::add_pointer_t<QPixmap>>(_a[1]))); break;
        case 2: _t->thumbnail3Changed((*reinterpret_cast<std::add_pointer_t<QPixmap>>(_a[1]))); break;
        case 3: _t->thumbnail4Changed((*reinterpret_cast<std::add_pointer_t<QPixmap>>(_a[1]))); break;
        case 4: _t->gotLogMessage((*reinterpret_cast<std::add_pointer_t<QString>>(_a[1]))); break;
        case 5: _t->gotStatusMessage((*reinterpret_cast<std::add_pointer_t<QString>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<uint>>(_a[2])),(*reinterpret_cast<std::add_pointer_t<uint>>(_a[3])),(*reinterpret_cast<std::add_pointer_t<bool>>(_a[4]))); break;
        case 6: _t->gotReloadPlaylists(); break;
        case 7: _t->gotReloadShaderParams(); break;
        case 8: _t->gotReloadCoreOptions(); break;
        case 9: _t->showErrorMessageDeferred((*reinterpret_cast<std::add_pointer_t<QString>>(_a[1]))); break;
        case 10: _t->showInfoMessageDeferred((*reinterpret_cast<std::add_pointer_t<QString>>(_a[1]))); break;
        case 11: _t->extractArchiveDeferred((*reinterpret_cast<std::add_pointer_t<QString>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<QString>>(_a[2])),(*reinterpret_cast<std::add_pointer_t<QString>>(_a[3])),(*reinterpret_cast<std::add_pointer_t<retro_task_callback_t>>(_a[4]))); break;
        case 12: _t->itemChanged(); break;
        case 13: _t->updateThumbnails(); break;
        case 14: _t->gridItemChanged((*reinterpret_cast<std::add_pointer_t<QString>>(_a[1]))); break;
        case 15: _t->gotThumbnailDownload((*reinterpret_cast<std::add_pointer_t<QString>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<QString>>(_a[2]))); break;
        case 16: _t->scrollToDownloads((*reinterpret_cast<std::add_pointer_t<QString>>(_a[1]))); break;
        case 17: _t->scrollToDownloadsAgain((*reinterpret_cast<std::add_pointer_t<QString>>(_a[1]))); break;
        case 18: _t->onBrowserDownloadsClicked(); break;
        case 19: _t->onBrowserUpClicked(); break;
        case 20: _t->onBrowserStartClicked(); break;
        case 21: _t->initContentTableWidget(); break;
        case 22: _t->onViewClosedDocksAboutToShow(); break;
        case 23: _t->onShowHiddenDockWidgetAction(); break;
        case 24: _t->setCoreActions(); break;
        case 25: _t->onRunClicked(); break;
        case 26: _t->loadContent((*reinterpret_cast<std::add_pointer_t<QHash<QString,QString>>>(_a[1]))); break;
        case 27: _t->onStartCoreClicked(); break;
        case 28: _t->onDropWidgetEnterPressed(); break;
        case 29: _t->selectBrowserDir((*reinterpret_cast<std::add_pointer_t<QString>>(_a[1]))); break;
        case 30: _t->setThumbnail((*reinterpret_cast<std::add_pointer_t<QString>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<QPixmap&>>(_a[2])),(*reinterpret_cast<std::add_pointer_t<bool>>(_a[3]))); break;
        case 31: _t->onResizeThumbnailOne((*reinterpret_cast<std::add_pointer_t<QPixmap&>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<bool>>(_a[2]))); break;
        case 32: _t->onResizeThumbnailTwo((*reinterpret_cast<std::add_pointer_t<QPixmap&>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<bool>>(_a[2]))); break;
        case 33: _t->onResizeThumbnailThree((*reinterpret_cast<std::add_pointer_t<QPixmap&>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<bool>>(_a[2]))); break;
        case 34: _t->onResizeThumbnailFour((*reinterpret_cast<std::add_pointer_t<QPixmap&>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<bool>>(_a[2]))); break;
        case 35: _t->appendLogMessage((*reinterpret_cast<std::add_pointer_t<QString>>(_a[1]))); break;
        case 36: _t->onGotLogMessage((*reinterpret_cast<std::add_pointer_t<QString>>(_a[1]))); break;
        case 37: _t->onGotStatusMessage((*reinterpret_cast<std::add_pointer_t<QString>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<uint>>(_a[2])),(*reinterpret_cast<std::add_pointer_t<uint>>(_a[3])),(*reinterpret_cast<std::add_pointer_t<bool>>(_a[4]))); break;
        case 38: _t->reloadPlaylists(); break;
        case 39: _t->deferReloadPlaylists(); break;
        case 40: _t->onGotReloadPlaylists(); break;
        case 41: _t->onGotReloadShaderParams(); break;
        case 42: _t->onGotReloadCoreOptions(); break;
        case 43: _t->showWelcomeScreen(); break;
        case 44: _t->onIconViewClicked(); break;
        case 45: _t->onListViewClicked(); break;
        case 46: _t->onBoxartThumbnailClicked(); break;
        case 47: _t->onScreenshotThumbnailClicked(); break;
        case 48: _t->onTitleThumbnailClicked(); break;
        case 49: _t->onLogoThumbnailClicked(); break;
        case 50: _t->onTabWidgetIndexChanged((*reinterpret_cast<std::add_pointer_t<int>>(_a[1]))); break;
        case 51: _t->deleteCurrentPlaylistItem(); break;
        case 52: _t->onFileDropWidgetContextMenuRequested((*reinterpret_cast<std::add_pointer_t<QPoint>>(_a[1]))); break;
        case 53: _t->showAbout(); break;
        case 54: _t->showDocs(); break;
        case 55: _t->onThumbnailPackExtractFinished((*reinterpret_cast<std::add_pointer_t<bool>>(_a[1]))); break;
        case 56: _t->deferReloadShaderParams(); break;
        case 57: _t->downloadThumbnail((*reinterpret_cast<std::add_pointer_t<QString>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<QString>>(_a[2])),(*reinterpret_cast<std::add_pointer_t<QUrl>>(_a[3]))); break;
        case 58: _t->downloadThumbnail((*reinterpret_cast<std::add_pointer_t<QString>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<QString>>(_a[2]))); break;
        case 59: _t->downloadAllThumbnails((*reinterpret_cast<std::add_pointer_t<QString>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<QUrl>>(_a[2]))); break;
        case 60: _t->downloadAllThumbnails((*reinterpret_cast<std::add_pointer_t<QString>>(_a[1]))); break;
        case 61: _t->downloadPlaylistThumbnails((*reinterpret_cast<std::add_pointer_t<QString>>(_a[1]))); break;
        case 62: _t->downloadNextPlaylistThumbnail((*reinterpret_cast<std::add_pointer_t<QString>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<QString>>(_a[2])),(*reinterpret_cast<std::add_pointer_t<QString>>(_a[3])),(*reinterpret_cast<std::add_pointer_t<QUrl>>(_a[4]))); break;
        case 63: _t->downloadNextPlaylistThumbnail((*reinterpret_cast<std::add_pointer_t<QString>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<QString>>(_a[2])),(*reinterpret_cast<std::add_pointer_t<QString>>(_a[3]))); break;
        case 64: _t->changeThumbnailType((*reinterpret_cast<std::add_pointer_t<ThumbnailType>>(_a[1]))); break;
        case 65: _t->onThumbnailDropped((*reinterpret_cast<std::add_pointer_t<QImage>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<ThumbnailType>>(_a[2]))); break;
        case 66: _t->onLoadCoreClicked((*reinterpret_cast<std::add_pointer_t<QStringList>>(_a[1]))); break;
        case 67: _t->onLoadCoreClicked(); break;
        case 68: _t->onUnloadCoreMenuAction(); break;
        case 69: _t->onTimeout(); break;
        case 70: _t->onCoreLoaded(); break;
        case 71: _t->onCurrentTableItemDataChanged((*reinterpret_cast<std::add_pointer_t<QModelIndex>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<QModelIndex>>(_a[2])),(*reinterpret_cast<std::add_pointer_t<QList<int>>>(_a[3]))); break;
        case 72: _t->onCurrentListItemChanged((*reinterpret_cast<std::add_pointer_t<QListWidgetItem*>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<QListWidgetItem*>>(_a[2]))); break;
        case 73: _t->onCurrentListItemDataChanged((*reinterpret_cast<std::add_pointer_t<QListWidgetItem*>>(_a[1]))); break;
        case 74: _t->onCurrentItemChanged((*reinterpret_cast<std::add_pointer_t<QModelIndex>>(_a[1]))); break;
        case 75: _t->onCurrentItemChanged((*reinterpret_cast<std::add_pointer_t<QHash<QString,QString>>>(_a[1]))); break;
        case 76: _t->onCurrentFileChanged((*reinterpret_cast<std::add_pointer_t<QModelIndex>>(_a[1]))); break;
        case 77: _t->onSearchEnterPressed(); break;
        case 78: _t->onSearchLineEditEdited((*reinterpret_cast<std::add_pointer_t<QString>>(_a[1]))); break;
        case 79: _t->onContentItemDoubleClicked((*reinterpret_cast<std::add_pointer_t<QModelIndex>>(_a[1]))); break;
        case 80: _t->onFileDoubleClicked((*reinterpret_cast<std::add_pointer_t<QModelIndex>>(_a[1]))); break;
        case 81: _t->onCoreLoadWindowClosed(); break;
        case 82: _t->onTreeViewItemsSelected((*reinterpret_cast<std::add_pointer_t<QModelIndexList>>(_a[1]))); break;
        case 83: _t->onSearchResetClicked(); break;
        case 84: _t->onLaunchWithComboBoxIndexChanged((*reinterpret_cast<std::add_pointer_t<int>>(_a[1]))); break;
        case 85: _t->onFileBrowserTreeContextMenuRequested((*reinterpret_cast<std::add_pointer_t<QPoint>>(_a[1]))); break;
        case 86: _t->onPlaylistWidgetContextMenuRequested((*reinterpret_cast<std::add_pointer_t<QPoint>>(_a[1]))); break;
        case 87: _t->onStopClicked(); break;
        case 88: _t->onZoomValueChanged((*reinterpret_cast<std::add_pointer_t<int>>(_a[1]))); break;
        case 89: _t->onPlaylistFilesDropped((*reinterpret_cast<std::add_pointer_t<QStringList>>(_a[1]))); break;
        case 90: _t->onShaderParamsClicked(); break;
        case 91: _t->onCoreOptionsClicked(); break;
        case 92: _t->onShowErrorMessage((*reinterpret_cast<std::add_pointer_t<QString>>(_a[1]))); break;
        case 93: _t->onShowInfoMessage((*reinterpret_cast<std::add_pointer_t<QString>>(_a[1]))); break;
        case 94: _t->onContributorsClicked(); break;
        case 95: _t->onItemChanged(); break;
        case 96: _t->onFileSystemDirLoaded((*reinterpret_cast<std::add_pointer_t<QString>>(_a[1]))); break;
        case 97: _t->onFileBrowserTableDirLoaded((*reinterpret_cast<std::add_pointer_t<QString>>(_a[1]))); break;
        case 98: _t->onDownloadScroll((*reinterpret_cast<std::add_pointer_t<QString>>(_a[1]))); break;
        case 99: _t->onDownloadScrollAgain((*reinterpret_cast<std::add_pointer_t<QString>>(_a[1]))); break;
        case 100: { int _r = _t->onExtractArchive((*reinterpret_cast<std::add_pointer_t<QString>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<QString>>(_a[2])),(*reinterpret_cast<std::add_pointer_t<QString>>(_a[3])),(*reinterpret_cast<std::add_pointer_t<retro_task_callback_t>>(_a[4])));
            if (_a[0]) *reinterpret_cast<int*>(_a[0]) = std::move(_r); }  break;
        case 101: _t->onThumbnailDownloadNetworkError((*reinterpret_cast<std::add_pointer_t<QNetworkReply::NetworkError>>(_a[1]))); break;
        case 102: _t->onThumbnailDownloadNetworkSslErrors((*reinterpret_cast<std::add_pointer_t<QList<QSslError>>>(_a[1]))); break;
        case 103: _t->onThumbnailDownloadFinished(); break;
        case 104: _t->onThumbnailDownloadProgress((*reinterpret_cast<std::add_pointer_t<qint64>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<qint64>>(_a[2]))); break;
        case 105: _t->onThumbnailDownloadReadyRead(); break;
        case 106: _t->onThumbnailDownloadCanceled(); break;
        case 107: _t->onDownloadThumbnail((*reinterpret_cast<std::add_pointer_t<QString>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<QString>>(_a[2]))); break;
        case 108: _t->onThumbnailPackDownloadNetworkError((*reinterpret_cast<std::add_pointer_t<QNetworkReply::NetworkError>>(_a[1]))); break;
        case 109: _t->onThumbnailPackDownloadNetworkSslErrors((*reinterpret_cast<std::add_pointer_t<QList<QSslError>>>(_a[1]))); break;
        case 110: _t->onThumbnailPackDownloadFinished(); break;
        case 111: _t->onThumbnailPackDownloadProgress((*reinterpret_cast<std::add_pointer_t<qint64>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<qint64>>(_a[2]))); break;
        case 112: _t->onThumbnailPackDownloadReadyRead(); break;
        case 113: _t->onThumbnailPackDownloadCanceled(); break;
        case 114: _t->onPlaylistThumbnailDownloadNetworkError((*reinterpret_cast<std::add_pointer_t<QNetworkReply::NetworkError>>(_a[1]))); break;
        case 115: _t->onPlaylistThumbnailDownloadNetworkSslErrors((*reinterpret_cast<std::add_pointer_t<QList<QSslError>>>(_a[1]))); break;
        case 116: _t->onPlaylistThumbnailDownloadFinished(); break;
        case 117: _t->onPlaylistThumbnailDownloadProgress((*reinterpret_cast<std::add_pointer_t<qint64>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<qint64>>(_a[2]))); break;
        case 118: _t->onPlaylistThumbnailDownloadReadyRead(); break;
        case 119: _t->onPlaylistThumbnailDownloadCanceled(); break;
        case 120: _t->startTimer(); break;
        case 121: _t->updateVisibleItems(); break;
        default: ;
        }
    }
    if (_c == QMetaObject::RegisterMethodArgumentMetaType) {
        switch (_id) {
        default: *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType(); break;
        case 71:
            switch (*reinterpret_cast<int*>(_a[1])) {
            default: *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType(); break;
            case 2:
                *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType::fromType< QList<int> >(); break;
            }
            break;
        }
    }
    if (_c == QMetaObject::IndexOfMethod) {
        if (QtMocHelpers::indexOfMethod<void (MainWindow::*)(const QPixmap & )>(_a, &MainWindow::thumbnailChanged, 0))
            return;
        if (QtMocHelpers::indexOfMethod<void (MainWindow::*)(const QPixmap & )>(_a, &MainWindow::thumbnail2Changed, 1))
            return;
        if (QtMocHelpers::indexOfMethod<void (MainWindow::*)(const QPixmap & )>(_a, &MainWindow::thumbnail3Changed, 2))
            return;
        if (QtMocHelpers::indexOfMethod<void (MainWindow::*)(const QPixmap & )>(_a, &MainWindow::thumbnail4Changed, 3))
            return;
        if (QtMocHelpers::indexOfMethod<void (MainWindow::*)(const QString & )>(_a, &MainWindow::gotLogMessage, 4))
            return;
        if (QtMocHelpers::indexOfMethod<void (MainWindow::*)(QString , unsigned  , unsigned  , bool )>(_a, &MainWindow::gotStatusMessage, 5))
            return;
        if (QtMocHelpers::indexOfMethod<void (MainWindow::*)()>(_a, &MainWindow::gotReloadPlaylists, 6))
            return;
        if (QtMocHelpers::indexOfMethod<void (MainWindow::*)()>(_a, &MainWindow::gotReloadShaderParams, 7))
            return;
        if (QtMocHelpers::indexOfMethod<void (MainWindow::*)()>(_a, &MainWindow::gotReloadCoreOptions, 8))
            return;
        if (QtMocHelpers::indexOfMethod<void (MainWindow::*)(QString )>(_a, &MainWindow::showErrorMessageDeferred, 9))
            return;
        if (QtMocHelpers::indexOfMethod<void (MainWindow::*)(QString )>(_a, &MainWindow::showInfoMessageDeferred, 10))
            return;
        if (QtMocHelpers::indexOfMethod<void (MainWindow::*)(QString , QString , QString , retro_task_callback_t )>(_a, &MainWindow::extractArchiveDeferred, 11))
            return;
        if (QtMocHelpers::indexOfMethod<void (MainWindow::*)()>(_a, &MainWindow::itemChanged, 12))
            return;
        if (QtMocHelpers::indexOfMethod<void (MainWindow::*)()>(_a, &MainWindow::updateThumbnails, 13))
            return;
        if (QtMocHelpers::indexOfMethod<void (MainWindow::*)(QString )>(_a, &MainWindow::gridItemChanged, 14))
            return;
        if (QtMocHelpers::indexOfMethod<void (MainWindow::*)(QString , QString )>(_a, &MainWindow::gotThumbnailDownload, 15))
            return;
        if (QtMocHelpers::indexOfMethod<void (MainWindow::*)(QString )>(_a, &MainWindow::scrollToDownloads, 16))
            return;
        if (QtMocHelpers::indexOfMethod<void (MainWindow::*)(QString )>(_a, &MainWindow::scrollToDownloadsAgain, 17))
            return;
    }
}

const QMetaObject *MainWindow::metaObject() const
{
    return QObject::d_ptr->metaObject ? QObject::d_ptr->dynamicMetaObject() : &staticMetaObject;
}

void *MainWindow::qt_metacast(const char *_clname)
{
    if (!_clname) return nullptr;
    if (!strcmp(_clname, qt_staticMetaObjectStaticContent<qt_meta_tag_ZN10MainWindowE_t>.strings))
        return static_cast<void*>(this);
    return QMainWindow::qt_metacast(_clname);
}

int MainWindow::qt_metacall(QMetaObject::Call _c, int _id, void **_a)
{
    _id = QMainWindow::qt_metacall(_c, _id, _a);
    if (_id < 0)
        return _id;
    if (_c == QMetaObject::InvokeMetaMethod) {
        if (_id < 122)
            qt_static_metacall(this, _c, _id, _a);
        _id -= 122;
    }
    if (_c == QMetaObject::RegisterMethodArgumentMetaType) {
        if (_id < 122)
            qt_static_metacall(this, _c, _id, _a);
        _id -= 122;
    }
    return _id;
}

// SIGNAL 0
void MainWindow::thumbnailChanged(const QPixmap & _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 0, nullptr, _t1);
}

// SIGNAL 1
void MainWindow::thumbnail2Changed(const QPixmap & _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 1, nullptr, _t1);
}

// SIGNAL 2
void MainWindow::thumbnail3Changed(const QPixmap & _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 2, nullptr, _t1);
}

// SIGNAL 3
void MainWindow::thumbnail4Changed(const QPixmap & _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 3, nullptr, _t1);
}

// SIGNAL 4
void MainWindow::gotLogMessage(const QString & _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 4, nullptr, _t1);
}

// SIGNAL 5
void MainWindow::gotStatusMessage(QString _t1, unsigned  _t2, unsigned  _t3, bool _t4)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 5, nullptr, _t1, _t2, _t3, _t4);
}

// SIGNAL 6
void MainWindow::gotReloadPlaylists()
{
    QMetaObject::activate(this, &staticMetaObject, 6, nullptr);
}

// SIGNAL 7
void MainWindow::gotReloadShaderParams()
{
    QMetaObject::activate(this, &staticMetaObject, 7, nullptr);
}

// SIGNAL 8
void MainWindow::gotReloadCoreOptions()
{
    QMetaObject::activate(this, &staticMetaObject, 8, nullptr);
}

// SIGNAL 9
void MainWindow::showErrorMessageDeferred(QString _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 9, nullptr, _t1);
}

// SIGNAL 10
void MainWindow::showInfoMessageDeferred(QString _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 10, nullptr, _t1);
}

// SIGNAL 11
void MainWindow::extractArchiveDeferred(QString _t1, QString _t2, QString _t3, retro_task_callback_t _t4)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 11, nullptr, _t1, _t2, _t3, _t4);
}

// SIGNAL 12
void MainWindow::itemChanged()
{
    QMetaObject::activate(this, &staticMetaObject, 12, nullptr);
}

// SIGNAL 13
void MainWindow::updateThumbnails()
{
    QMetaObject::activate(this, &staticMetaObject, 13, nullptr);
}

// SIGNAL 14
void MainWindow::gridItemChanged(QString _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 14, nullptr, _t1);
}

// SIGNAL 15
void MainWindow::gotThumbnailDownload(QString _t1, QString _t2)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 15, nullptr, _t1, _t2);
}

// SIGNAL 16
void MainWindow::scrollToDownloads(QString _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 16, nullptr, _t1);
}

// SIGNAL 17
void MainWindow::scrollToDownloadsAgain(QString _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 17, nullptr, _t1);
}
QT_WARNING_POP
